import factory
from factory.django import DjangoModelFactory

from tiers.models import Tier
from fake_organizations.models import Organization


class OrganizationFactory(DjangoModelFactory):
    class Meta(object):
        model = Organization

    name = factory.Sequence('organization name {}'.format)


class TierFactory(DjangoModelFactory):
    class Meta(object):
        model = Tier

    organization = factory.SubFactory(OrganizationFactory)



"""
dummy functions for instrumentation

fall back to these if `beeline` isn't available.
"""
import functools


def add_context_field(name, value):
    pass


def traced(name, trace_id=None, parent_id=None):
    def decorator_traced(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        return wrapper
    return decorator_traced



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
from django.conf import settings
import django.utils.timezone
import model_utils.fields
import tiers.models

from ..app_settings import ORGANIZATION_MODEL


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(ORGANIZATION_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Tier',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('name', models.CharField(default=b'trial', max_length=255, choices=[(b'trial', b'Trial')])),
                ('tier_enforcement_exempt', models.BooleanField(default=False)),
                ('tier_enforcement_grace_period', models.PositiveIntegerField(default=14)),
                ('tier_expires_at', models.DateTimeField(default=tiers.models.set_default_expiration)),
                ('organization', models.OneToOneField(related_name='tier', to=ORGANIZATION_MODEL, on_delete=models.deletion.DO_NOTHING)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import logging

from django.shortcuts import redirect
from django.utils.deprecation import MiddlewareMixin
from django.urls import NoReverseMatch, reverse

from .models import Tier
from .app_settings import EXPIRED_REDIRECT_URL, ORGANIZATION_TIER_GETTER_NAME

try:
    import beeline
except ImportError:
    import tiers.instrumentation as beeline

log = logging.getLogger(__name__)


class TierMiddleware(MiddlewareMixin):
    """
    Django Tiers middleware
    """
    @beeline.traced(name="TiersMiddleware.process_request")
    def process_request(self, request):
        """
        Fetch organization from session and deny access to the system if the tier
        is expired
        """
        # If we're aleady on the url where we have to be, do nothing
        if (EXPIRED_REDIRECT_URL is not None) and (request.path.rstrip('/') in EXPIRED_REDIRECT_URL.rstrip('/')):
            beeline.add_context_field("tiers.no_action_required", True)
            return

        # try/catch needed because the URLs don't necessarily exist both in AMC and edX
        try:
            # If we're trying to log out or release a hijacked user, don't redirect to expired page
            if request.path == reverse("release_hijack") or request.path == reverse("account_logout"):
                return
        except NoReverseMatch:
            pass

        # Nothing to do if the user is not logged in
        if not request.user.is_authenticated:
            return

        # If the user has superuser privileges don't do anything
        if request.user.is_superuser:
            return

        # If there is no organization in the sesssion fail silenty.
        # This should not happen.
        if not request.session.get('organization'):
            beeline.add_context_field("tiers.no_organization", True)
            return

        org = request.session['organization']
        beeline.add_context_field("tiers.organization", "{}".format(org))
        try:
            if ORGANIZATION_TIER_GETTER_NAME:
                tier = org.__getattribute__(ORGANIZATION_TIER_GETTER_NAME)()
            else:
                tier = org.tier
        except Exception:
            # If the organization for some reason does not have a tier assigned
            # fail silently. This should not happen. We should always automatically create
            # a tier for each organization.
            beeline.add_context_field("tiers.organization_without_tier", True)
            log.warning("Organization wihout Tier: {0}".format(org))
            return

        # Only display expiration warning for Trial tiers for now
        request.session['DISPLAY_EXPIRATION_WARNING'] = ((tier.name == Tier.TIERS.TRIAL) and
                                                         (not tier.tier_enforcement_exempt))
        request.session['TIER_EXPIRES_IN'] = tier.time_til_tier_expires()
        beeline.add_context_field("tiers.tier_expires_in", request.session['TIER_EXPIRES_IN'])
        # TODO: I'm not sure if we have to refresh the session info at this point somehow.
        request.session['TIER_EXPIRED'] = tier.has_tier_expired()
        beeline.add_context_field("tiers.tier_expired", request.session['TIER_EXPIRED'])
        # TODO: We should use request.TIER_NAME instead of meddling the session, but being consistent for now
        request.session['TIER_NAME'] = tier.name
        beeline.add_context_field("tiers.tier_name", tier.name)

        # TODO: I'm not sure if we have to refresh the session info at this point somehow.
        if tier.has_tier_expired():
            if EXPIRED_REDIRECT_URL is None:
                return
            else:
                return redirect(EXPIRED_REDIRECT_URL)



from django.conf import settings

ORGANIZATION_MODEL = getattr(settings, 'TIERS_ORGANIZATION_MODEL', 'organizations.Organization')
EXPIRED_REDIRECT_URL = getattr(settings, 'TIERS_EXPIRED_REDIRECT_URL', None)

ORGANIZATION_TIER_GETTER_NAME = getattr(settings, 'TIERS_ORGANIZATION_TIER_GETTER_NAME', None)



from __future__ import unicode_literals

import re

from datetime import datetime, timedelta
from django.contrib.auth.models import User
from django.test.client import RequestFactory
from mock import patch, Mock

from tiers.tests.utils import TiersTestCaseBase
from tiers.tests.factories import TierFactory
from tiers.middleware import TierMiddleware


class TiersTests(TiersTestCaseBase):

    def test_non_expired_tier(self):
        t = TierFactory()
        assert t.tier_enforcement_exempt is False
        assert t.has_tier_expired() is False
        assert t.has_tier_grace_period_expired() is False

    def test_expired_tier(self):
        t = TierFactory(tier_expires_at=(datetime.now() - timedelta(days=2)))
        assert t.tier_enforcement_exempt is False
        assert t.has_tier_expired() is True
        assert t.has_tier_grace_period_expired() is False

    def test_expired_grade_period(self):
        t = TierFactory(tier_expires_at=(datetime.now() - timedelta(days=20)))
        assert t.tier_enforcement_exempt is False
        assert t.has_tier_expired() is True
        assert t.has_tier_grace_period_expired() is True

    def test_exemption(self):
        t = TierFactory(
            tier_enforcement_exempt=True,
            tier_expires_at=(datetime.now() - timedelta(days=20)))
        assert t.tier_enforcement_exempt is True
        assert t.has_tier_expired() is False
        assert t.has_tier_grace_period_expired() is False

    def test_expire_message(self):
        """
        Ensures that `has_tier_grace_period_expired` works well.

        This function uses regexps match what the Django `timeuntil` returns.
        The `avoid_wrapping` function has more details in the docstring.
        """
        t = TierFactory.create(tier_expires_at=datetime(2019, 2, 5) + timedelta(days=30))
        assert t.tier_expires_at == datetime(2019, 3, 7)

        message = t.time_til_tier_expires(now=datetime(2019, 2, 6))
        assert re.match('4.*weeks,.*1.*day', message), 'Should handle long time well'

        message = t.time_til_tier_expires(now=datetime(2019, 3, 4, 0, 8, 0))
        assert re.match(r'2.*days,.*23.*hours', message), 'Should short time well'

        message = t.time_til_tier_expires(now=datetime(2020, 2, 6))
        assert re.match('0.*minutes', message), 'Should handle future time well.'


@patch.object(User, 'is_authenticated', Mock(return_value=True))
@patch('tiers.middleware.reverse', Mock(return_value='/something'))
class TestMiddlewareTests(TiersTestCaseBase):
    def setUp(self):
        super(TestMiddlewareTests, self).setUp()
        self.middleware = TierMiddleware()
        self.tier = TierFactory.create(tier_expires_at=datetime.now()+timedelta(days=40))
        self.request = RequestFactory().get('/dashboard')
        self.request.session = {}
        self.request.session['organization'] = self.tier.organization
        self.user = User()
        self.request.user = self.user

    def test_empty_by_default_attributes(self):
        default = object()
        for attrib in ['DISPLAY_EXPIRATION_WARNING', 'TIER_EXPIRES_IN', 'TIER_EXPIRED', 'TIER_NAME']:
            assert self.request.session.get(attrib, default) is default

    def test_added_session_attribs(self):
        self.middleware.process_request(self.request)
        assert not self.request.session['TIER_EXPIRED']
        assert self.request.session['TIER_EXPIRES_IN'] == self.tier.time_til_tier_expires()
        assert self.request.session['DISPLAY_EXPIRATION_WARNING']
        assert self.request.session['TIER_NAME'] == 'trial'
